(* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre *)

(*s Interpretation. *)

open Sml_ng

(*s Root contexts and space contexts. *)

let root_context = 
  { b = false; em = false; i = false; s = false; tt = false;
    u = 0; size = -1 (* any *); color = Any (* any *) }

let space_context ctx =
  { ctx with s = false; em = false; i = false; b = false;
      color = if ctx.u == 0 then White else ctx.color }

(*s The interpretation of a sequence of non-space characters is the stream
    of these characters, all decorated with the current context [ctx]. *)

let interp_string ctx s =
  let n = String.length s in
  let rec interp i =
    if i == n then [< >] else [< '(s.[i],ctx); interp (succ i) >]
  in
  interp 0

(*s Interpretation of a space. *)

let char_of_space = function
  | SPC -> ' '
  | TAB -> '\t'
  | LF  -> '\n'
  | CR  -> '\r'

(*s The interpretation of a document is computed by the recursive functions
    [interp] (for an item) and [interp_many] (for a list of items). 
    [ctx] is the current context.
    [lsp] is the last space context, of type [context option]; if [None],
    the previous outut character was not a space; if [Some c], it was
    a space with context [c]. *)

let interp_document d =
  let rec interp lsp ctx = function
    | Text s -> 
	interp_string ctx s, None
    | Wspace w -> 
	let sp_ctx = space_context ctx in
	let c = char_of_space w in
	if ctx.tt then
	  [< '(c, sp_ctx) >], None
	else begin match lsp with
	  | Some sp_ctx' when sp_ctx' = sp_ctx ->
	      [< >], lsp
	  | _ ->
	      [< ' (' ', sp_ctx) >], (Some sp_ctx)
	end
    | Tag (B,d) ->
	interp_many lsp { ctx with b = true } d
    | Tag (EM,d) ->
	interp_many lsp (if ctx.s then ctx else { ctx with em = not ctx.em }) d
    | Tag (I,d) ->
	interp_many lsp { ctx with i = true } d
    | Tag (PL,d) ->
	interp_many lsp { ctx with u = 0; s = false; em = false; i = false;
                         	b = false; tt = false } d
    | Tag (S,d) ->
	interp_many lsp { ctx with s = true; em = false } d
    | Tag (TT,d) ->
	interp_many lsp { ctx with tt = true } d
    | Tag (U,d) ->
	interp_many 
	  lsp (if ctx.u < 3 then {ctx with u = succ ctx.u} else ctx) d
    | Tag ((Size n),d) ->
	interp_many lsp { ctx with size = n } d
    | Tag ((Color c),d) ->
	interp_many lsp { ctx with color = c } d
  and interp_many lsp ctx = function
    | [] -> 
	[< >], lsp
    | item :: items -> 
	let st, lsp' = interp lsp ctx item in
	let st', lsp'' = interp_many lsp' ctx items in
	[< st; st' >], lsp''
  in
  fst (interp_many None root_context d)
