(* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre *)

(*s Set the [Alarm] signal according to the given time limit. 
    5 seconds should be enough to output the solution. *)

let tlimit = int_of_string Sys.argv.(1)
let _ = Unix.alarm (tlimit - 5)

(*s Read the initial document from stdin, store it in [init_doc], 
    and output it on timeout. *)

let init_doc = Parse.from_stdin ()
let timeout _ = Output.output_document init_doc; flush stdout; exit 0
let _ = Sys.set_signal Sys.sigalrm (Sys.Signal_handle timeout)

(*s Main program. We compute first the interpretation [id] of the document,
    then the optimized document [od] from it; if it is smaller than
    the original document, we output it; otherwise we output the original
    document. *)

let main () =
  let id = Interp.interp_document init_doc in
  let od = Optim.optim_document id in
  let old_size = Output.size init_doc in
  let new_size = Optim.size od in
  if new_size < old_size then begin
    ignore (Sys.set_signal Sys.sigalrm Sys.Signal_ignore);
    Optim.output_optimized_document od
  end else
    Output.output_document init_doc

let _ =
  Printexc.catch main ()



