/* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre */

/*s Yacc grammar for SML/NG documents. Immediate. One remark, yet:
    when checking that the closing tag matched the opening one, we
    use physical equality since tags are statically allocated in the
    lexer. */

%{

open Sml_ng

%}

%token <Sml_ng.tag> OPEN
%token <Sml_ng.tag> CLOSE
%token <string> TEXT
%token <Sml_ng.white_space> WSP
%token EOF

%type <Sml_ng.document> document
%start document

%%

document:
  items EOF { $1 }
;

items:
  /* epsilon */  { [] }
| item items     { $1 :: $2 }
;

item:
  TEXT 
    { Text $1 }
| WSP  
    { Wspace $1 }
| OPEN items CLOSE 
    { let t1 = $1 and t2 = $3 in
      if t1 != t2 then raise Parsing.Parse_error;
      Tag (t1, $2) }
;

  
